@@ -29,6 +29,7 @@
 #include "TMethodArg.h"
 #include "TROOT.h"
 #include "TSystem.h"
+#include "TThread.h"
 
 // Standard
 #include <assert.h>
@@ -120,7 +121,10 @@ CallWrapper* new_CallWrapper(CallWrapper::DeclId_t fid, const std::string& n)
 }
 
 typedef std::vector<TGlobal*> GlobalVars_t;
+typedef std::map<TGlobal*, GlobalVars_t::size_type> GlobalVarsIndices_t;
+
 static GlobalVars_t g_globalvars;
+static GlobalVarsIndices_t g_globalidx;
 
 
 // data ----------------------------------------------------------------------
@@ -216,6 +220,9 @@ public:
     // macro that resolves to the ::CppyyLegacy::GetROOT() function call)
         (void)gROOT;
 
+    // create the Cling interpreter lock
+        TThread::Initialize();
+
     // setup dummy holders for global and std namespaces
         assert(g_classrefs.size() == GLOBAL_HANDLE);
         g_name2classrefidx[""]      = GLOBAL_HANDLE;
@@ -228,6 +235,7 @@ public:
 
     // add a dummy global to refer to as null at index 0
         g_globalvars.push_back(nullptr);
+        g_globalidx[nullptr] = 0;
 
     // fill out the builtins
         std::set<std::string> bi{g_builtins};
@@ -428,6 +436,15 @@ std::string Cppyy::ResolveName(const std::string& cppitem_name)
         tclean.replace(pos, 4, "::");
         pos += 2;
     }
+
+#ifdef _WIN32
+// TODO: this is a bug somewhere deep in Clang's type printing; since we're close
+// to upgrading to Clang9, this hack will have to do.
+    pos = tclean.find("std::basic_string<char,std::allocator<char>,std::allocator<char> >");
+    if (pos != std::string::npos)
+        tclean = tclean.substr(0, pos+28) + "char_traits" + tclean.substr(pos+37, std::string::npos);
+#endif
+
     if (tclean.compare(0, 6, "const ") != 0)
         return TClassEdit::ShortType(tclean.c_str(), 2);
     return "const " + TClassEdit::ShortType(tclean.c_str(), 2);
@@ -926,6 +943,7 @@ Cppyy::TCppObject_t Cppyy::CallO(TCppMethod_t method,
     void* obj = ::operator new(gInterpreter->ClassInfo_Size(cr->GetClassInfo()));
     if (WrapperCall(method, nargs, args, self, obj))
         return (TCppObject_t)obj;
+    ::operator delete(obj);
     return (TCppObject_t)0;
 }
 
@@ -989,6 +1007,7 @@ bool Cppyy::IsEnum(const std::string& type_name)
 // variable declarations; full resolution does, but simply removing it will do:
     if (tn_short.rfind("enum ", 0) == 0)
         tn_short = tn_short.substr(5, std::string::npos);
+    R__LOCKGUARD_CLING(gInterpreterMutex);
     return gInterpreter->ClassInfo_IsEnum(tn_short.c_str());
 }
 
@@ -1194,7 +1213,7 @@ std::vector<Cppyy::TCppScope_t> Cppyy::GetUsingNamespaces(TCppScope_t scope)
 
     const std::vector<std::string>& v = gInterpreter->GetUsingNamespaces(cr->GetClassInfo());
     res.reserve(v.size());
-    for (auto uid : v) {
+    for (const auto& uid : v) {
         Cppyy::TCppScope_t uscope = GetScope(uid);
         if (uscope) res.push_back(uscope);
     }
@@ -1285,7 +1304,9 @@ bool Cppyy::IsSubtype(TCppType_t derived, TCppType_t base)
         return true;
     TClassRef& derived_type = type_from_handle(derived);
     TClassRef& base_type = type_from_handle(base);
-    return derived_type->GetBaseClass(base_type) != 0;
+    if (derived_type.GetClass() && base_type.GetClass())
+        return derived_type->GetBaseClass(base_type) != 0;
+    return false;
 }
 
 bool Cppyy::IsSmartPtr(TCppType_t klass)
@@ -1394,7 +1415,7 @@ Cppyy::TCppIndex_t Cppyy::GetNumMethods(TCppScope_t scope, bool accept_namespace
             // instantiation, so force it
                 std::ostringstream stmt;
                 stmt << "template class " << clName << ";";
-                gInterpreter->Declare(stmt.str().c_str());
+                gInterpreter->Declare(stmt.str().c_str(), true /* silent */);
 
             // now reload the methods
                 return (TCppIndex_t)cr->GetListOfMethods(true)->GetSize();
@@ -1801,8 +1822,11 @@ std::string type_remap(const std::string& n1, const std::string& n2)
         if (n2 == "std::basic_string<wchar_t,std::char_traits<wchar_t>,std::allocator<wchar_t> >")
             return n2;                      // match like for like
         return "std::string";               // probably best bet
-    } else if (n1 == "float")
+    } else if (n1 == "float") {
         return "double";                    // debatable, but probably intended
+    } else if (n1 == "complex") {
+        return "std::complex<double>";
+    }
     return n1;
 }
 
@@ -1998,6 +2022,21 @@ intptr_t Cppyy::GetDatamemberOffset(TCppScope_t scope, TCppIndex_t idata)
     return (intptr_t)-1;
 }
 
+static inline
+Cppyy::TCppIndex_t gb2idx(TGlobal* gb)
+{
+    if (!gb) return (Cppyy::TCppIndex_t)-1;
+
+    auto pidx = g_globalidx.find(gb);
+    if (pidx == g_globalidx.end()) {
+        auto idx = g_globalvars.size();
+        g_globalvars.push_back(gb);
+        g_globalidx[gb] = idx;
+        return (Cppyy::TCppIndex_t)idx;
+    }
+    return (Cppyy::TCppIndex_t)pidx->second;
+}
+
 Cppyy::TCppIndex_t Cppyy::GetDatamemberIndex(TCppScope_t scope, const std::string& name)
 {
     if (scope == GLOBAL_HANDLE) {
@@ -2029,11 +2068,7 @@ Cppyy::TCppIndex_t Cppyy::GetDatamemberIndex(TCppScope_t scope, const std::strin
             if (wrap && wrap->GetAddress()) gb = wrap;
         }
 
-        if (gb) {
-        // TODO: do we ever need a reverse lookup?
-            g_globalvars.push_back(gb);
-            return TCppIndex_t(g_globalvars.size() - 1);
-        }
+        return gb2idx(gb);
 
     } else {
         TClassRef& cr = type_from_handle(scope);
@@ -2048,6 +2083,16 @@ Cppyy::TCppIndex_t Cppyy::GetDatamemberIndex(TCppScope_t scope, const std::strin
     return (TCppIndex_t)-1;
 }
 
+Cppyy::TCppIndex_t Cppyy::GetDatamemberIndexEnumerated(TCppScope_t scope, TCppIndex_t idata)
+{
+    if (scope == GLOBAL_HANDLE) {
+        TGlobal* gb = (TGlobal*)((THashList*)gROOT->GetListOfGlobals(false /* load */))->At((int)idata);
+        return gb2idx(gb);
+    }
+
+    return idata;
+}
+
 
 // data member properties ----------------------------------------------------
 bool Cppyy::IsPublicData(TCppScope_t scope, TCppIndex_t idata)
@@ -2669,6 +2714,9 @@ int cppyy_datamember_index(cppyy_scope_t scope, const char* name) {
     return (int)Cppyy::GetDatamemberIndex(scope, name);
 }
 
+int cppyy_datamember_index_enumerated(cppyy_scope_t scope, int datamember_index) {
+    return (int)Cppyy::GetDatamemberIndexEnumerated(scope, datamember_index);
+}
 
 
 /* data member properties ------------------------------------------------- */
